/**********************************************************************/
/* exe_cube                                                           */
/*                                                                    */
/* Drawing a balanced sample                                          */
/*                                                                    */
/* Author : Guillaume Chauvet <chauvet@ensai.fr>                      */
/* Version 1.0                                                        */
/* Copyright LGPL 2.1 <http://www.gnu.org/copyleft/lesser.html>       */
/*                                                                    */
/* exe_cube Drawing a balanced sample                                 */
/* Copyright (C) 2005 Guillaume Chauvet                               */
/*                                                                    */
/* CREST - Laboratoire de Statistique d'Enqutes                      */
/* Ensai                                                              */
/* Campus de Ker Lann                                                 */
/* 35170 Bruz                                                         */
/*                                                                    */
/* E-mail: chauvet@ensai.fr                                           */
/*                                                                    */
/**********************************************************************/
/* This library is free software; you can redistribute it and/or      */
/* modify it under the terms of the GNU Lesser General Public         */
/* License as published by the Free Software Foundation; either       */
/* version 1.0 of the License, or (at your option) any later version. */
/*                                                                    */
/* This library is distributed in the hope that it will be useful,    */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of     */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  */
/* Lesser General Public License for more details.                    */
/*                                                                    */
/* You should have received a copy of the GNU Lesser General Public   */
/* License along with this library; if not, write to the Free Software*/ 
/*    Foundation, Inc., 59 Temple Place, Suite 330, Boston,           */
/*    MA  02111-1307  USA                                             */
/**********************************************************************/

/***********************************************************************/
/* The exe_cube macro enables to draw a sample with the Fast Cube      */
/* method                                                              */
/*    base : library containing the data base                          */
/*    data : data base                                                 */
/*    id : variable identifying the units of the population            */
/*    pi : variable of the inclusion probabilities                     */
/*    contr : names of the balancing variables, spaced with blanks     */
/*    sort : name of the SAS output table, containing the same         */
/*           variables as data, and a variable named ech equalling 1 if*/
/*           the unit belongs to the sample and 0 otherwise            */
/*    atter : landing option (1,2 or 3)                                */
/*    compeq : a strictly nonnegative real value if we want the        */
/*             complementary of the sample to be balanced on the same  */
/*             variables, and 0 otherwise (default value : 0)          */
/*                                                                     */
/* Comment :                                                           */
/*    The option "compeq" is useless if equal inclusion probabilities  */
/*    are used                                                         */
/*                                                                     */
/* The exe_cube macro uses SAS/MACRO and SAS/IML modules               */
/*                                                                     */
/* The exe_cube macro uses the sub-macros :                            */
/*    vol                                                              */
/*    atterrissage1                                                    */
/*    atterrissage2                                                    */
/*    atterrissage3                                                    */
/*    modules_IML                                                      */
/*    (See below)                                                      */ 
/***********************************************************************/;

%macro exe_cube(base=,data=,id=,pi=,contr=,sort=,atter=,compeq=0);

%let contr2=&contr;

data &sort;set &base..&data;ech=&pi;
%if &compeq>0 %then %do;
	%let i=1;%let comp=;
	%do %while (%scan(&contr,&i) ne);
		%let piv=%scan(&contr,&i)_comp;
		&piv=%scan(&contr,&i)/(1-&pi);
		%let comp=&comp &piv; 
		%let i=%eval(&i+1);
	%end;
	%let contr2=&contr &comp;
%end;
run;

proc iml;

%modules_IML;

use &sort where (&pi>0 & &pi<1);
read all var {&id} into _id;
read all var {&contr2} into _contr_;
read all var {&pi} into _pi;

%vol(_pi=_pi,_contr_=_contr_,_pistar=_pistar);
ech=round(_pistar);

call symput("harry",char(ncol(loc(_pistar>1e-8 & _pistar<1-1e-8))));

%if &harry>0 %then %do;

	%if &atter=1 %then %do; 
		%atterrissage1(_pi=_pi,_pistar=_pistar,_contr_=_contr_,_pifou=_pifou); 
	%end;
	%if &atter=2 %then %do; 
		%atterrissage2(_pi=_pi,_pistar=_pistar,_contr_=_contr_,_pifou=_pifou); 
	%end;
	%if &atter=3 %then %do; 
		%atterrissage3(_pi=_pi,_pistar=_pistar,_contr_=_contr_,_pifou=_pifou); 
	%end;

	ech=round(_pifou);

%end;

edit &sort var {ech};replace all var {ech} where (ech>0 & ech<1);close &sort;  

quit;

%mend exe_cube;

/**********************************************************************/
/* vol                                                                */
/*                                                                    */
/* Performing a flight phase                                          */
/*                                                                    */
/* Author : Guillaume Chauvet <chauvet@ensai.fr>                      */
/* Version 1.0                                                        */
/* Copyright LGPL 2.1 <http://www.gnu.org/copyleft/lesser.html>       */
/*                                                                    */
/* vol Performing a flight phase                                      */
/* Copyright (C) 2005 Guillaume Chauvet                               */
/*                                                                    */
/* CREST - Laboratoire de Statistique d'Enqutes                      */
/* Ensai                                                              */
/* Campus de Ker Lann                                                 */
/* 35170 Bruz                                                         */
/*                                                                    */
/* E-mail: chauvet@ensai.fr                                           */
/*                                                                    */
/**********************************************************************/
/* This library is free software; you can redistribute it and/or      */
/* modify it under the terms of the GNU Lesser General Public         */
/* License as published by the Free Software Foundation; either       */
/* version 1.0 of the License, or (at your option) any later version. */
/*                                                                    */
/* This library is distributed in the hope that it will be useful,    */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of     */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  */
/* Lesser General Public License for more details.                    */
/*                                                                    */
/* You should have received a copy of the GNU Lesser General Public   */
/* License along with this library; if not, write to the Free Software*/ 
/*    Foundation, Inc., 59 Temple Place, Suite 330, Boston,           */
/*    MA  02111-1307  USA                                             */
/**********************************************************************/

/**********************************************************************/
/* The vol macro performs a flight phase                              */
/* Input :                                                            */
/*   -> pi : vector of inclusion probabilities                        */
/*   -> contr : matrix of balancing variables                         */
/* Output :                                                           */
/*   -> pistar : vector of inclusion probabilities after the flight   */
/*               phase                                                */
/*                                                                    */
/* The vol macro uses SAS/MACRO and SAS/IML modules                   */
/*                                                                    */
/* The vol macro uses the sub-macro : modules_IML (see below)         */
/**********************************************************************/;

%macro vol(_pi=,_contr_=,_pistar=);

/* Initialization of the parameters                                   */

N=nrow(&_contr_);
p=ncol(&_contr_);
k=1;
_phi=&_pi;
&_pistar=&_pi;

/* Fast flight phase                             */

if p+1<=N then do;

	_phi=&_pi[1:(p+1)];
	_r=t(1:p+1);
	_B_=(_phi##(-1))#&_contr_[1:p+1,];
	k=p+2;

	/* A first iteration                                              */

	run choix_vecteur(_B_,_phi);
	drop_=loc(_phi<1e-8 | _phi>1-1e-8);
	r=ncol(drop_);
	
	/* As long as units for which the choice has been made can        */
	/* be replaced, we go on                                          */

	do while(r<=N-k+1);

		&_pistar[_r[drop_]]=_phi[drop_];
		_r[drop_]=t(k:k+r-1);
		_phi[drop_]=&_pi[k:k+r-1];

		_B_[drop_,]=((_phi[drop_])##(-1))#&_contr_[_r[drop_],];
		k=k+r;

		run choix_vecteur(_B_,_phi);

		drop_=loc(_phi<1e-8 | _phi>1-1e-8);
		r=ncol(drop_);

	end;

	keep_=loc(_phi>1e-8 & 1-_phi>1e-8);
	&_pistar[_r[drop_]]=_phi[drop_];

end;

/* Complement of the flight phase                                     */

if (ncol(keep_)>0)|(k<=N) then do;

	/* We join the remaining units together                           */

	if ncol(keep_)=0 then do;
		keep_={1};
		_r=t(k:N);
		_phi=&_pi[k:N];
		_B_=((&_pi[k:N])##(-1))#&_contr_[k:N,]; 
	end;
	else do;
		if k>N then do;
			_r=_r[keep_];	
			_phi=_phi[keep_];
			_B_=_B_[keep_,];
		end;
		else do;
			_r=_r[keep_]//t(k:N);	
			_phi=_phi[keep_]//&_pi[k:N];
			_B_=_B_[keep_,]//(((&_pi[k:N])##(-1))#&_contr_[k:N,]); 
		end;
	end;

	rank=rang(_B_);
	
	do while ((rank<nrow(_B_))&(ncol(keep_)>0));

		run choix_vecteur(_B_,_phi);

		drop_=loc(_phi<=1e-8 | 1-_phi<=1e-8);
		keep_=loc(_phi>1e-8 & 1-_phi>1e-8);
		&_pistar[_r[drop_]]=_phi[drop_];
		if ncol(keep_)>0 then do;
				_r=_r[keep_];
				_phi=_phi[keep_];
				_B_=_B_[keep_,];
			end;
	
	end;

end;

if ncol(keep_)>0 then &_pistar[_r]=_phi;

%mend vol;

/**********************************************************************/
/* atterrissage1                                                      */
/*                                                                    */
/* Performing a landing phase by successively giving up constraints   */
/*                                                                    */
/* Author : Guillaume Chauvet <chauvet@ensai.fr>                      */
/* Version 1.0                                                        */
/* Copyright LGPL 2.1 <http://www.gnu.org/copyleft/lesser.html>       */
/*                                                                    */
/* atterrissage1 Performing a landing phase by successively giving up */
/* constraints                                                        */
/* Copyright (C) 2005 Guillaume Chauvet                               */
/*                                                                    */
/* CREST - Laboratoire de Statistique d'Enqutes                      */
/* Ensai                                                              */
/* Campus de Ker Lann                                                 */
/* 35170 Bruz                                                         */
/*                                                                    */
/* E-mail: chauvet@ensai.fr                                           */
/*                                                                    */
/**********************************************************************/
/* This library is free software; you can redistribute it and/or      */
/* modify it under the terms of the GNU Lesser General Public         */
/* License as published by the Free Software Foundation; either       */
/* version 1.0 of the License, or (at your option) any later version. */
/*                                                                    */
/* This library is distributed in the hope that it will be useful,    */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of     */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  */
/* Lesser General Public License for more details.                    */
/*                                                                    */
/* You should have received a copy of the GNU Lesser General Public   */
/* License along with this library; if not, write to the Free Software*/ 
/*    Foundation, Inc., 59 Temple Place, Suite 330, Boston,           */
/*    MA  02111-1307  USA                                             */
/**********************************************************************/

/**********************************************************************/
/* The atterrissage1 macro performs a landing phase by successively   */
/* giving up constraints                                              */
/* Input :                                                            */
/*   -> pi : vector of inclusion probabilities before the flight      */
/*           phase                                                    */
/*   -> pistar : vector of inclusion probabilities after the flight   */
/*               phase                                                */
/*   -> contr : matrix of the balancing variables                     */
/* Output :                                                           */
/*   -> pifou : vector indicating the belonging to the sample         */
/*                                                                    */
/* Can only be used if pistar has some noninteger components          */
/*                                                                    */
/* The atterrissage1 macro uses SAS/MACRO and SAS/IML modules         */
/*                                                                    */
/* The atterrissage1 macro uses the sub-macro : modules_IML           */
/* (see below)                                                        */
/**********************************************************************/;

%macro atterrissage1(_pi=,_pistar=,_contr_=,_pifou=);

_r=t(loc(&_pistar>1e-8 & &_pistar<1-1e-8));
_phi=&_pistar[_r];
_psi=&_pi[_r];
_B_=(_psi##(-1))#&_contr_[_r,];
&_pifou=&_pistar;

nrow=nrow(_B_);
do while (nrow>1);

	_B_=_B_[,1:ncol(_B_)-1];
	rank=rang(_B_);
	nrow=nrow(_B_);
	ncol=ncol(_r);

	do while ((rank<nrow)&(ncol>0));

		run choix_vecteur(_B_,_phi);
		drop_=loc(^(_phi>1e-8 & 1-_phi>1e-8));
		keep_=loc(_phi>1e-8 & 1-_phi>1e-8);
		&_pifou[_r[drop_]]=_phi[drop_];
		if ncol(keep_)>0 then do; 
			_r=_r[keep_];_phi=_phi[keep_];_B_=_B_[keep_,];
			nrow=nrow(_B_);ncol=ncol(_r);
		end;
		else do; 
			nrow=0;ncol=0;
		end;
	
	end;

end;

if (ncol(keep_)>0) then &_pifou[_r]=(uniform(0)<_phi);

%mend atterrissage1;

/**********************************************************************/
/* atterrissage2                                                      */
/*                                                                    */
/* Performing a landing phase by minimizing a gap to the balancing    */
/*                                                                    */
/* Author : Guillaume Chauvet <chauvet@ensai.fr>                      */
/* Version 1.0                                                        */
/* Copyright LGPL 2.1 <http://www.gnu.org/copyleft/lesser.html>       */
/*                                                                    */
/* atterrissage2 Performing a landing phase by minimizing a gap to the*/
/* balancing                                                          */
/* Copyright (C) 2005 Guillaume Chauvet                               */
/*                                                                    */
/* CREST - Laboratoire de Statistique d'Enqutes                      */
/* Ensai                                                              */
/* Campus de Ker Lann                                                 */
/* 35170 Bruz                                                         */
/*                                                                    */
/* E-mail: chauvet@ensai.fr                                           */
/*                                                                    */
/**********************************************************************/
/* This library is free software; you can redistribute it and/or      */
/* modify it under the terms of the GNU Lesser General Public         */
/* License as published by the Free Software Foundation; either       */
/* version 1.0 of the License, or (at your option) any later version. */
/*                                                                    */
/* This library is distributed in the hope that it will be useful,    */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of     */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  */
/* Lesser General Public License for more details.                    */
/*                                                                    */
/* You should have received a copy of the GNU Lesser General Public   */
/* License along with this library; if not, write to the Free Software*/ 
/*    Foundation, Inc., 59 Temple Place, Suite 330, Boston,           */
/*    MA  02111-1307  USA                                             */
/**********************************************************************/

/**********************************************************************/
/* The atterrissage2 macro performs a landing phase by considering    */
/* all the possible samples among the remaining units, and selecting  */
/* preferably those giving a low difference to the balancing          */
/* Input :                                                            */
/*   -> pi : vector of inclusion probabilities before the flight      */
/*           phase                                                    */
/*   -> pistar : vector of inclusion probabilities after the flight   */
/*               phase                                                */
/*   -> contr : matrix of the balancing variables                     */
/* Output :                                                           */
/*   -> pifou : vector indicating the belonging to the sample         */
/*                                                                    */
/* Can only be used if pistar has some noninteger components          */
/*                                                                    */
/* The atterrissage2 macro uses SAS/MACRO and SAS/IML modules         */
/*                                                                    */
/* The atterrissage2 macro uses the sub-macro : modules_IML           */
/* (see below)                                                        */
/**********************************************************************/;

%macro atterrissage2(_pi=,_pistar=,_contr_=,_pifou=);

_r=t(loc(&_pistar>1e-8 & &_pistar<1-1e-8));
_phi=&_pistar[_r];
_psi=&_pi[_r];
_B_=(_psi##(-1))#&_contr_[_r,];

/* Building of the table used for the linear optimisation (use of  */
/* the LP proc)                                                    */
/* Definition of the first column                                  */ 

nr=nrow(_r);
piv1=repeat("cons",nr+1,1);piv2=left(char(t(1:nr+1)));
_row_={"obj"}//concat(piv1,piv2)//{"lowerb"};
varnames={_row_};
create _row_ from _row_ [colname=varnames];append from _row_;close _row_;

/* Definition of the coefficients of the issue (calculation of     */
/* the costs)                                                      */

tot_=&_contr_[+,];
_s=J(nr,1,0);
_cout_=((t(_s-_phi)*_B_)/tot_)[##]//_s;	
do while (_s[+]<nr);
	j=loc(_s=0)[><];
	_s[1:j]=1-_s[1:j];
	_cout_=_cout_||(((t(_s-_phi)*_B_)/tot_)[##]//_s);	
end;
piv1=J(1,2##nr,1);piv2=J(1,2##nr,0);
_cout_=_cout_//piv1//piv2;
varnames=concat(repeat("x",1,2##nr),left(char(1:2##nr)));
create _cout_ from _cout_ [colname=varnames];append from _cout_;close _cout_;

/* Definition of the last but one column, changed into a SAS table */

_type_={"min"}//repeat("eq",nr+1,1)//{"lowerbd"};
varnames={_type_};
create _type_ from _type_ [colname=varnames];append from _type_;close _type_;

/* Definition of the last column, changed into a SAS table         */
_rhs_={.}//_phi//{1,.};
varnames={_rhs_};
create _rhs_ from _rhs_ [colname=varnames];append from _rhs_;close _rhs_;

/* Vectors &_pistar and _r must be available                        */
create &_pistar from &_pistar;append from &_pistar;close &_pistar;
create _r from _r;append from _r;close _r;
quit;

/* Solution of the optimisation problem                            */ 
data lptable;merge _row_ _cout_ _type_ _rhs_;run;
proc lp data=lptable primalout=suppmin;run;
/* Back into SAS IML                                               */ 
proc iml;

use suppmin;read all var {_value_} where (_TYPE_="LOWERBD") into _patter;
use &_pistar;read all into &_pistar;
use _r;read all into _r;
	
/* Initialization of the final vector giving the sampling outcome  */ 
&_pifou=&_pistar;

/* Selection of a sample according to the optimal sampling design  */
_support_=loc(abs(_patter)>1e-8);
_support_=t(_support_)||cusum(_patter[_support_]);
alea=uniform(0);
s=_support_[loc(_support_[,2]>alea)[1],1];
s=s-1;
_s_=J(nrow(_r),1,0);
i=1;
do while (s>1); _s_[i]=mod(s,2); s=int(s/2); i=i+1; end;
_s_[i]=s;
&_pifou[_r]=_s_;
	
%mend atterrissage2;

/**********************************************************************/
/* atterrissage3                                                      */
/*                                                                    */
/* Performing a landing phase by minimizing a gap to the balancing    */
/*                                                                    */
/* Author : Guillaume Chauvet <chauvet@ensai.fr>                      */
/* Version 1.0                                                        */
/* Copyright LGPL 2.1 <http://www.gnu.org/copyleft/lesser.html>       */
/*                                                                    */
/* atterrissage3 Performing a landing phase by minimizing a gap to the*/
/* balancing                                                          */
/* Copyright (C) 2005 Guillaume Chauvet                               */
/*                                                                    */
/* CREST - Laboratoire de Statistique d'Enqutes                      */
/* Ensai                                                              */
/* Campus de Ker Lann                                                 */
/* 35170 Bruz                                                         */
/*                                                                    */
/* E-mail: chauvet@ensai.fr                                           */
/*                                                                    */
/**********************************************************************/
/* This library is free software; you can redistribute it and/or      */
/* modify it under the terms of the GNU Lesser General Public         */
/* License as published by the Free Software Foundation; either       */
/* version 1.0 of the License, or (at your option) any later version. */
/*                                                                    */
/* This library is distributed in the hope that it will be useful,    */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of     */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  */
/* Lesser General Public License for more details.                    */
/*                                                                    */
/* You should have received a copy of the GNU Lesser General Public   */
/* License along with this library; if not, write to the Free Software*/ 
/*    Foundation, Inc., 59 Temple Place, Suite 330, Boston,           */
/*    MA  02111-1307  USA                                             */
/**********************************************************************/

/**********************************************************************/
/* The atterrissage3 macro works like the preceeding one, but only    */
/* considering the samples whose size equals the sum of inclusion     */
/* probabilities. We obtain a fixed sample size. If this option is    */
/* used, the variable of inclusion probabilities must have been used  */
/* for the flight phase                                               */
/* Input :                                                            */
/*   -> pi : vector of inclusion probabilities before the flight      */
/*           phase                                                    */
/*   -> pistar : vector of inclusion probabilities after the flight   */
/*               phase                                                */
/*   -> contr : matrix of the balancing variables                     */
/* Output :                                                           */
/*   -> pifou : vector indicating the belonging to the sample         */
/*                                                                    */
/* Can only be used if pistar has some noninteger components          */
/*                                                                    */
/* The atterrissage3 macro uses SAS/MACRO and SAS/IML modules         */
/*                                                                    */
/* The atterrissage3 macro uses the sub-macro : modules_IML           */
/* (see below)                                                        */
/**********************************************************************/;

%macro atterrissage3(_pi=,_pistar=,_contr_=,_pifou=);

_r=t(loc(&_pistar>1e-8 & &_pistar<1-1e-8));
_phi=&_pistar[_r];
_psi=&_pi[_r];
_B_=(_psi##(-1))#&_contr_[_r,];

/* Building of the table used for the linear optimisation (use of  */
/* the LP proc)                                                    */
/* Definition of the first column                                  */ 

nr=nrow(_r);
piv1=repeat("cons",nr+1,1);piv2=left(char(t(1:nr+1)));
_row_={"obj"}//concat(piv1,piv2)//{"lowerb"};
varnames={_row_};
create _row_ from _row_ [colname=varnames];append from _row_;close _row_;

/* We round off the number of units remaining to sample, using the */
/* splitting method (Deville and Till, 1998)                      */  

nphi=floor(_phi[+]);
alpha=_phi[+]-nphi;

if (alpha ^= 0) then do;

	_phi1=_phi;
	_phi2=_phi;
	yop=1;
	suma=0;
	do while (suma < alpha);
		a=min(_phi[yop],alpha-suma,alpha*(1-_phi[yop])/(1-alpha));
		_phi1[yop]=_phi[yop]-a;
		_phi2[yop]=_phi[yop]+(1/alpha-1)*a;
		suma=suma+a;
		yop=yop+1;
	end;

	_phi=_phi1+(uniform(0)<alpha)*(_phi2-_phi1);

end;

/* Definition of the coefficients of the issue (calculation of     */
/* the costs)                                                      */
tot_=&_contr_[+,];

natter=round(_phi[+]); 
_s=J(natter,1,1)//J(nr-natter,1,0); 

_cout_=((t(_s-_phi)*_B_)/tot_)[##]//_s;	
do while (_s[nr-natter+1:nr][+] < natter);
	j=loc(_s=0)[><];
	_s[1:j]=1-_s[1:j];
	if _s[+]=natter then do;
		_cout_=_cout_||(((t(_s-_phi)*_B_)/tot_)[##]//_s);	
	end;
end;
nbech=fact(nr)/(fact(natter)*fact(nr-natter));
piv1=J(1,nbech,1);piv2=J(1,nbech,0);
_cout_=_cout_//piv1//piv2;
varnames=concat(repeat("x",1,nbech),left(char(1:nbech)));
create _cout_ from _cout_ [colname=varnames];append from _cout_;close _cout_;

/* Definition of the last but one column, changed into a SAS table */
_type_={"min"}//repeat("eq",nr+1,1)//{"lowerbd"};
varnames={_type_};
create _type_ from _type_ [colname=varnames];append from _type_;close _type_;

/* Definition of the last column, changed into a SAS table         */
_rhs_={.}//_phi//{1,.};
varnames={_rhs_};
create _rhs_ from _rhs_ [colname=varnames];append from _rhs_;close _rhs_;

/* Vectors &_pistar and _r must be available                        */
create &_pistar from &_pistar;append from &_pistar;close &_pistar;
create _r from _r;append from _r;close _r;
quit;

/* Solution of the optimisation problem                            */ 
data lptable;merge _row_ _cout_ _type_ _rhs_;run;
proc lp data=lptable primalout=suppmin;run;
/* Back into SAS IML                                               */ 
proc iml;

use suppmin;read all var {_value_} where (_TYPE_="LOWERBD") into _patter;
use &_pistar;read all into &_pistar;
use _r;read all into _r;
use _cout_;read all into _cout_;

/* Initialization of the final vector giving the sampling outcome  */ 
&_pifou=&_pistar;

/* Selection of a fixed size sample according to the optimal       */
/* sampling design                                                 */
_support_=loc(abs(_patter)>1e-8);
_support_=t(_support_)||cusum(_patter[_support_]);
alea=uniform(0);
s=_support_[loc(_support_[,2]>alea)[1],1];
_s_=_cout_[2:nrow(_cout_)-2,s];
&_pifou[_r]=_s_;

%mend atterrissage3;

%macro modules_IML;

	/******************************************************************/
	/* Basic loop of the flight phase                                 */
	/******************************************************************/

	start choix_vecteur(_mat_,_pistar_);
		_u_=(homogen(t(_mat_)))[,1];
		_nonnul_=loc(abs(_u_)>1e-8);
		piv=_u_[_nonnul_]; 
		_lem_=(piv##(-1))#((piv>1e-8)-_pistar_[_nonnul_]);
		l1=min(_lem_);
		_lem_=(piv##(-1))#(_pistar_[_nonnul_]-(piv<1e-8));
		l2=min(_lem_);
		if (uniform(0)<(l2/(l1+l2))) then _pistar_=_pistar_+_u_*l1;
		else _pistar_=_pistar_-_u_*l2;
	finish choix_vecteur;

	/******************************************************************/
	/* Calculation of the rank of a matrix                            */
	/******************************************************************/
	start rang(_mat_);
		X=homogen(block(_mat_,0));
		return (nrow(X)-ncol(X));
	finish rang;

%mend modules_IML;

